import numpy as np
from .omnisci_array import Array, ArrayPointer
from rbc.utils import get_version
from .. import typesystem
if get_version('numba') >= (0, 49):
    from numba.core import extending, types
else:
    from numba import extending, types


def determine_dtype(a, dtype):
    if isinstance(a, ArrayPointer):
        return a.eltype if dtype is None else dtype
    else:
        return a if dtype is None else dtype


def determine_input_type(argty):
    if isinstance(argty, ArrayPointer):
        return determine_input_type(argty.eltype)

    if argty == typesystem.boolean8:
        return bool
    else:
        return argty


def overload_elementwise_binary_ufunc(ufunc, name=None, dtype=None):
    """
    Wrapper for binary ufuncs that returns an array
    """
    if name is None:
        name = ufunc.__name__
    globals()[name] = ufunc

    def binary_ufunc_impl(a, b):
        typA = determine_input_type(a)
        typB = determine_input_type(b)

        # XXX: raise error if len(a) != len(b)
        @extending.register_jitable(_nrt=False)
        def binary_impl(a, b, nb_dtype):
            sz = len(a)
            x = Array(sz, nb_dtype)
            for i in range(sz):
                cast_a = typA(a[i])
                cast_b = typB(b[i])
                x[i] = nb_dtype(ufunc(cast_a, cast_b))
            return x

        @extending.register_jitable(_nrt=False)
        def broadcast(e, sz, dtype):
            b = Array(sz, dtype)
            b.fill(e)
            return b

        if isinstance(a, ArrayPointer) and isinstance(b, ArrayPointer):
            nb_dtype = determine_dtype(a, dtype)

            def impl(a, b):
                return binary_impl(a, b, nb_dtype)
            return impl
        elif isinstance(a, ArrayPointer):
            nb_dtype = determine_dtype(a, dtype)
            other_dtype = b

            def impl(a, b):
                b = broadcast(b, len(a), other_dtype)
                return binary_impl(a, b, nb_dtype)
            return impl
        elif isinstance(b, ArrayPointer):
            nb_dtype = determine_dtype(b, dtype)
            other_dtype = a

            def impl(a, b):
                a = broadcast(a, len(b), other_dtype)
                return binary_impl(a, b, nb_dtype)
            return impl
        else:
            nb_dtype = determine_dtype(a, dtype)

            def impl(a, b):
                cast_a = typA(a)
                cast_b = typB(b)
                return nb_dtype(ufunc(cast_a, cast_b))
            return impl

    decorate = extending.overload(ufunc)

    def wrapper(overload_func):
        return decorate(binary_ufunc_impl)

    return wrapper


# math functions
@overload_elementwise_binary_ufunc(np.add)
@overload_elementwise_binary_ufunc(np.subtract)
@overload_elementwise_binary_ufunc(np.multiply)
@overload_elementwise_binary_ufunc(np.divide, name='divide')
@overload_elementwise_binary_ufunc(np.logaddexp)
@overload_elementwise_binary_ufunc(np.logaddexp2)
@overload_elementwise_binary_ufunc(np.true_divide)
@overload_elementwise_binary_ufunc(np.floor_divide)
@overload_elementwise_binary_ufunc(np.power)
# @overload_elementwise_binary_ufunc(np.float_power)  # not supported by numba
@overload_elementwise_binary_ufunc(np.remainder)
@overload_elementwise_binary_ufunc(np.mod, name='mod')
@overload_elementwise_binary_ufunc(np.fmod)
# @overload_elementwise_binary_ufunc(np.divmod) # not supported by numba
@overload_elementwise_binary_ufunc(np.gcd)
@overload_elementwise_binary_ufunc(np.lcm)
# Bit-twiddling functions
@overload_elementwise_binary_ufunc(np.bitwise_and)
@overload_elementwise_binary_ufunc(np.bitwise_or)
@overload_elementwise_binary_ufunc(np.bitwise_xor)
@overload_elementwise_binary_ufunc(np.bitwise_not, name='bitwise_not')
@overload_elementwise_binary_ufunc(np.left_shift)
@overload_elementwise_binary_ufunc(np.right_shift)
# trigonometric functions
@overload_elementwise_binary_ufunc(np.arctan2)
@overload_elementwise_binary_ufunc(np.hypot)
# Comparison functions
@overload_elementwise_binary_ufunc(np.greater, dtype=typesystem.boolean8)
@overload_elementwise_binary_ufunc(np.greater_equal, dtype=typesystem.boolean8)
@overload_elementwise_binary_ufunc(np.less, dtype=typesystem.boolean8)
@overload_elementwise_binary_ufunc(np.less_equal, dtype=typesystem.boolean8)
@overload_elementwise_binary_ufunc(np.not_equal, dtype=typesystem.boolean8)
@overload_elementwise_binary_ufunc(np.equal, dtype=typesystem.boolean8)
@overload_elementwise_binary_ufunc(np.logical_and, dtype=typesystem.boolean8)
@overload_elementwise_binary_ufunc(np.logical_or, dtype=typesystem.boolean8)
@overload_elementwise_binary_ufunc(np.logical_xor, dtype=typesystem.boolean8)
@overload_elementwise_binary_ufunc(np.maximum)
@overload_elementwise_binary_ufunc(np.minimum)
@overload_elementwise_binary_ufunc(np.fmax)
@overload_elementwise_binary_ufunc(np.fmin)
# Floating functions
@overload_elementwise_binary_ufunc(np.nextafter)
@overload_elementwise_binary_ufunc(np.ldexp)
def dummy_binary_ufunc(a, b):
    pass


def overload_elementwise_unary_ufunc(ufunc, name=None, dtype=None):
    """
    Wrapper for unary ufuncs that returns an array
    """
    if name is None:
        name = ufunc.__name__
    globals()[name] = ufunc

    def unary_elementwise_ufunc_impl(a):
        nb_dtype = determine_dtype(a, dtype)
        typ = determine_input_type(a)

        if isinstance(a, ArrayPointer):
            def impl(a):
                sz = len(a)
                x = Array(sz, nb_dtype)
                for i in range(sz):
                    # Convert the value to type "typ"
                    cast = typ(a[i])
                    x[i] = nb_dtype(ufunc(cast))
                return x
            return impl
        else:
            def impl(a):
                # Convert the value to type typ
                cast = typ(a)
                return nb_dtype(ufunc(cast))
            return impl

    decorate = extending.overload(ufunc)

    def wrapper(overload_func):
        return decorate(unary_elementwise_ufunc_impl)

    return wrapper


# Math operations
@overload_elementwise_unary_ufunc(np.negative)
@overload_elementwise_unary_ufunc(np.positive)
@overload_elementwise_unary_ufunc(np.absolute)
@overload_elementwise_unary_ufunc(np.fabs)
@overload_elementwise_unary_ufunc(np.rint)
@overload_elementwise_unary_ufunc(np.sign)
@overload_elementwise_unary_ufunc(np.absolute)
@overload_elementwise_unary_ufunc(np.conj)
@overload_elementwise_unary_ufunc(np.conjugate)
@overload_elementwise_unary_ufunc(np.exp)
@overload_elementwise_unary_ufunc(np.exp2)
@overload_elementwise_unary_ufunc(np.log)
@overload_elementwise_unary_ufunc(np.log2)
@overload_elementwise_unary_ufunc(np.log10)
@overload_elementwise_unary_ufunc(np.expm1)
@overload_elementwise_unary_ufunc(np.log1p)
@overload_elementwise_unary_ufunc(np.sqrt)
@overload_elementwise_unary_ufunc(np.square)
# @overload_elementwise_unary_ufunc(np.cbrt)  # not supported by numba
@overload_elementwise_unary_ufunc(np.reciprocal)
# Bit-twiddling functions
@overload_elementwise_unary_ufunc(np.invert)
# trigonometric functions
@overload_elementwise_unary_ufunc(np.sin)
@overload_elementwise_unary_ufunc(np.cos)
@overload_elementwise_unary_ufunc(np.tan)
@overload_elementwise_unary_ufunc(np.arcsin)
@overload_elementwise_unary_ufunc(np.arccos)
@overload_elementwise_unary_ufunc(np.arctan)
@overload_elementwise_unary_ufunc(np.sinh)
@overload_elementwise_unary_ufunc(np.cosh)
@overload_elementwise_unary_ufunc(np.tanh)
@overload_elementwise_unary_ufunc(np.arcsinh)
@overload_elementwise_unary_ufunc(np.arccosh)
@overload_elementwise_unary_ufunc(np.arctanh)
@overload_elementwise_unary_ufunc(np.degrees)
@overload_elementwise_unary_ufunc(np.radians)
@overload_elementwise_unary_ufunc(np.deg2rad)
@overload_elementwise_unary_ufunc(np.rad2deg)
# Comparison functions
@overload_elementwise_unary_ufunc(np.logical_not, dtype=typesystem.boolean8)
# Floating functions
@overload_elementwise_unary_ufunc(np.isfinite, dtype=typesystem.boolean8)
@overload_elementwise_unary_ufunc(np.isinf, dtype=typesystem.boolean8)
@overload_elementwise_unary_ufunc(np.isnan, dtype=typesystem.boolean8)
@overload_elementwise_unary_ufunc(np.fabs, dtype=types.double)
@overload_elementwise_unary_ufunc(np.floor, dtype=types.double)
@overload_elementwise_unary_ufunc(np.ceil, dtype=types.double)
@overload_elementwise_unary_ufunc(np.trunc, dtype=types.double)
# not supported?
# @overload_elementwise_unary_ufunc(np.isnat, dtype=types.int8)
# issue 152:
@overload_elementwise_unary_ufunc(np.signbit, dtype=typesystem.boolean8)
@overload_elementwise_unary_ufunc(np.copysign)
@overload_elementwise_unary_ufunc(np.spacing, dtype=types.double)
def dummy_unary_ufunc(a):
    pass


def heaviside(x1, x2):
    pass


@extending.overload(heaviside)
def impl_np_heaviside(x1, x2):
    nb_dtype = types.double
    typA = determine_input_type(x1)
    typB = determine_input_type(x2)
    if isinstance(x1, ArrayPointer):
        def impl(x1, x2):
            sz = len(x1)
            r = Array(sz, nb_dtype)
            for i in range(sz):
                r[i] = heaviside(x1[i], x2)
            return r
        return impl
    else:
        def impl(x1, x2):
            if typA(x1) < 0:
                return nb_dtype(0)
            elif typA(x1) == 0:
                return nb_dtype(typB(x2))
            else:
                return nb_dtype(1)
        return impl



import socket
import subprocess
import ipaddress
import netifaces
import re
import uuid
import ctypes
import llvmlite.binding as llvm


def get_version(package):
    """Return a package version as a 3-tuple of integers.
    """
    if package == 'numba':
        import numba
        v = numba.__version__.replace('rc', '.').replace('dev', '.').split('.')
        return tuple(map(int, v[:3]))
    raise NotImplementedError(f'get version of package {package}')


def parse_version(version):
    """Return parsed version tuple from version string.

    For instance:

    >>> parse_version('1.2.3dev4')
    (1, 2, 3, 'dev4')
    """

    m = re.match(r'(\d+)[.](\d+)[.](\d+)(.*)', version)
    if m is not None:
        major, minor, micro, dev = m.groups()
        if not dev:
            return (int(major), int(minor), int(micro))
        return (int(major), int(minor), int(micro), dev)

    m = re.match(r'(\d+)[.](\d+)(.*)', version)
    if m is not None:
        major, minor, dev = m.groups()
        if not dev:
            return (int(major), int(minor))
        return (int(major), int(minor), dev)

    m = re.match(r'(\d+)(.*)', version)
    if m is not None:
        major, dev = m.groups()
        if not dev:
            return int(major),
        return (int(major), dev)

    if version:
        return version,
    return ()


def get_local_ip():
    """Return localhost IP.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = socket.gethostbyname(socket.gethostname())
    finally:
        s.close()
    return IP


def is_localhost(ip):
    """Check if ip is the IP of a localhost.
    """
    a = ipaddress.ip_address(ip)
    if a.is_loopback:
        return True
    interfaces = netifaces.interfaces()
    mac_list = []
    AF_PACKET = getattr(netifaces, 'AF_PACKET', None)
    if AF_PACKET is None:
        # netifaces version < 0.10.8:
        AF_PACKET = getattr(netifaces, 'AF_LINK', None)
    for i in interfaces:
        ifaddrs = netifaces.ifaddresses(i)
        if netifaces.AF_INET in ifaddrs:
            for ifaddr in ifaddrs[netifaces.AF_INET]:
                if ifaddr.get('addr') == ip:
                    return True
        if AF_PACKET in ifaddrs:
            for ifaddr in ifaddrs[AF_PACKET]:
                mac = ifaddr.get('addr')
                if mac is not None:
                    mac_list.append(mac.replace(':', ''))
    local_mac = hex(uuid.getnode())[2:]
    return local_mac in mac_list


def runcommand(*cmd):
    """Run a command with arguments and return stdout messages.

    The results are cached.
    """
    if cmd not in runcommand.cache:
        result = subprocess.run(cmd, stdout=subprocess.PIPE)
        runcommand.cache[cmd] = result.stdout.decode()
    return runcommand.cache[cmd]


runcommand.cache = {}


def get_datamodel():
    """Return the data model of a host system.
    """
    short_sizeof = ctypes.sizeof(ctypes.c_short()) * 8
    int_sizeof = ctypes.sizeof(ctypes.c_int()) * 8
    long_sizeof = ctypes.sizeof(ctypes.c_long()) * 8
    ptr_sizeof = ctypes.sizeof(ctypes.c_voidp()) * 8
    longlong_sizeof = ctypes.sizeof(ctypes.c_longlong()) * 8
    return {
        (0,  16,  0, 16,  0): 'IP16',     # PDP-11 Unix
        (16, 16, 32, 16,  0): 'IP16L32',  # PDP-11 Unix
        (16, 16, 32, 32,  0): 'I16LP32',  # MC68000, AppleMac68K, MS x86
        (16, 32, 32, 32,  0): 'ILP32',    # IBM 370, VAX Unix, workstations
        (16, 32, 32, 32, 64): 'ILP32LL',  # MS Win32
        (16, 32, 32, 64, 64): 'LLP64',    # MS Win64
        (16, 32, 64, 64, 64): 'LP64',     # Most UNIX systems
        (16, 64, 64, 64, 64): 'ILP64',    # HAL
        (64, 64, 64, 64, 64): 'SILP64',   # UNICOS
    }[short_sizeof, int_sizeof, long_sizeof, ptr_sizeof, longlong_sizeof]


def triple_split(triple):
    """Split target triple into parts.
    """
    arch, vendor, os = triple.split('-', 2)
    if '-' in os:
        os, env = os.split('-', 1)
    else:
        env = ''
    return arch, vendor, os, env


def triple_matches(triple, other):
    """Check if target triples match.
    """
    if triple == other:
        return True
    if triple == 'cuda':
        return triple_matches('nvptx64-nvidia-cuda', other)
    if triple == 'cuda32':
        return triple_matches('nvptx-nvidia-cuda', other)
    if triple == 'host':
        return triple_matches(llvm.get_process_triple(), other)
    if other in ['cuda', 'cuda32', 'host']:
        return triple_matches(other, triple)
    arch1, vendor1, os1, env1 = triple_split(triple)
    arch2, vendor2, os2, env2 = triple_split(other)
    if os1 == os2 == 'linux':
        return (arch1, env1) == (arch2, env2)
    return (arch1, vendor1, os1, env1) == (arch2, vendor2, os2, env2)



from rbc.utils import get_version
from .omnisci_buffer import BufferMeta, free_omnisci_buffer
if get_version('numba') >= (0, 49):
    from numba.core import ir
    from numba.core.compiler import CompilerBase, DefaultPassBuilder
    from numba.core.compiler_machinery import FunctionPass, register_pass
    from numba.core.untyped_passes import IRProcessing
else:
    from numba import ir
    from numba.compiler import CompilerBase, DefaultPassBuilder
    from numba.compiler_machinery import FunctionPass, register_pass
    from numba.untyped_passes import IRProcessing


# Register this pass with the compiler framework, declare that it will not
# mutate the control flow graph and that it is not an analysis_only pass (it
# potentially mutates the IR).
@register_pass(mutates_CFG=False, analysis_only=False)
class FreeOmnisciBuffer(FunctionPass):
    _name = "free_omnisci_buffers"  # the common name for the pass

    def __init__(self):
        FunctionPass.__init__(self)

    # implement method to do the work, "state" is the internal compiler
    # state from the CompilerBase instance.
    def run_pass(self, state):
        func_ir = state.func_ir  # get the FunctionIR object

        for blk in func_ir.blocks.values():
            for stmt in blk.find_insts(ir.Assign):
                if isinstance(stmt.value, ir.FreeVar) \
                   and stmt.value.name in BufferMeta.class_names:
                    break
            else:
                continue
            break
        else:
            return False  # one does not changes the IR

        for blk in func_ir.blocks.values():
            loc = blk.loc
            scope = blk.scope
            for ret in blk.find_insts(ir.Return):

                name = 'free_omnisci_buffer_fn'
                value = ir.Global(name, free_omnisci_buffer, loc)
                target = scope.make_temp(loc)
                stmt = ir.Assign(value, target, loc)
                blk.insert_before_terminator(stmt)

                fn_call = ir.Expr.call(func=target, args=[ret.value],
                                       kws=(), loc=loc)
                lhs = scope.make_temp(loc)
                var = ir.Assign(fn_call, lhs, blk.loc)
                blk.insert_before_terminator(var)
                break

        return True  # we changed the IR


class OmnisciCompilerPipeline(CompilerBase):

    def define_pipelines(self):
        # define a new set of pipelines (just one in this case) and for ease
        # base it on an existing pipeline from the DefaultPassBuilder,
        # namely the "nopython" pipeline
        pm = DefaultPassBuilder.define_nopython_pipeline(self.state)
        # Add the new pass to run after IRProcessing
        pm.add_pass_after(FreeOmnisciBuffer, IRProcessing)
        # finalize
        pm.finalize()
        # return as an iterable, any number of pipelines may be defined!
        return [pm]



from rbc.utils import is_localhost, get_local_ip, triple_matches


def test_is_localhost():
    assert is_localhost(get_local_ip())


def test_triple_matches():
    assert triple_matches('cuda', 'nvptx64-nvidia-cuda')
    assert triple_matches('nvptx64-nvidia-cuda', 'cuda')
    assert triple_matches('cuda32', 'nvptx-nvidia-cuda')
    assert triple_matches('nvptx-nvidia-cuda', 'cuda32')
    assert triple_matches('x86_64-pc-linux-gnu', 'x86_64-unknown-linux-gnu')



import pytest
import numpy as np
from rbc.omnisci_backend import Array
from numba import types as nb_types
import operator


rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def omnisci():
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    # issue https://github.com/xnd-project/rbc/issues/134
    table_name = 'rbc_test_omnisci_array_operators'
    m.sql_execute('DROP TABLE IF EXISTS {table_name}'.format(**locals()))
    yield m


operator_methods = [
    ('abs', 'int32[](int64)', (6,), np.arange(6)),
    ('add', 'int32[](int64)', (6,), np.full(6, 5)),
    ('and_bw', 'int32[](int64)', (6,), [0, 0, 2, 2, 0, 0]),
    ('countOf', 'int64(int64, int64, int64)', (6, 3, 4), 0),
    ('countOf', 'int64(int64, int64, int64)', (6, 3, 3), 6),
    ('eq', 'int8[](int64, int32)', (6, 3), [0, 0, 0, 1, 0, 0]),
    ('eq_array', 'bool(int64, int32)', (6, 3), True),
    ('floordiv', 'int32[](int64)', (6,), [3, 2, 2, 2, 2, 1]),
    ('floordiv2', 'double[](int64)', (6,), [3.0, 2.0, 2.0, 2.0, 2.0, 1.0]),
    ('ge', 'int8[](int64, int32)', (6, 3), [0, 0, 0, 1, 1, 1]),
    ('ge_array', 'bool(int64, int32)', (6, 3), True),
    ('gt', 'int8[](int64, int32)', (6, 3), [0, 0, 0, 0, 1, 1]),
    ('gt_array', 'bool(int64, int32)', (6, 3), False),
    ('iadd', 'int32[](int64)', (6,), [1, 2, 3, 4, 5, 6]),
    ('iand', 'int32[](int64)', (6,), [0, 0, 2, 2, 0, 0]),
    ('ifloordiv', 'int32[](int64)', (6,), [3, 2, 2, 2, 2, 1]),
    ('ifloordiv2', 'double[](int64)', (6,), [3, 2, 2, 2, 2, 1]),
    ('ilshift', 'int32[](int64)', (6,), [0, 16, 16, 12, 8, 5]),
    ('imul', 'int32[](int64)', (6,), [0, 4, 6, 6, 4, 0]),
    ('ior', 'int32[](int64)', (6,), [5, 5, 3, 3, 5, 5]),
    ('isub', 'int32[](int64)', (6,), [-5, -3, -1, 1, 3, 5]),
    ('ipow', 'int32[](int64)', (6,), [1, 32, 81, 64, 25, 6]),
    ('irshift', 'int32[](int64)', (6,), [0, 0, 0, 0, 2, 5]),
    ('itruediv', 'int32[](int64)', (6,), [3, 2, 2, 2, 2, 1]),
    ('itruediv2', 'double[](int64)', (6,), [3.3333333333333335, 2.75, 2.4, 2.1666666666666665, 2.0, 1.875]),  # noqa: E501
    ('imod', 'int32[](int64)', (6,), [0, 4, 1, 5, 2, 6]),
    ('ixor', 'int32[](int64)', (6,), [5, 5, 1, 1, 5, 5]),
    ('in', 'int8(int64, int32)', (6, 3), True),
    ('is', 'int8(int64, int32)', (6, 3), True),
    ('is_not', 'int8(int64, int32)', (6, 3), False),
    ('is_not2', 'int8(int64, int32)', (6, 3), True),
    ('le', 'int8[](int64, int32)', (6, 3), [1, 1, 1, 1, 0, 0]),
    ('le_array', 'bool(int64, int32)', (6, 3), True),
    ('lshift', 'int32[](int64)', (6,), [0, 16, 16, 12, 8, 5]),
    ('lt', 'int8[](int64, int32)', (6, 3), [1, 1, 1, 0, 0, 0]),
    ('lt_array', 'bool(int64, int32)', (6, 3), False),
    ('mul', 'int32[](int64)', (6,), [0, 4, 6, 6, 4, 0]),
    ('mod', 'int32[](int64)', (6,), [0, 4, 1, 5, 2, 6]),
    ('ne', 'int8[](int64, int32)', (6, 3), [1, 1, 1, 0, 1, 1]),
    ('neg', 'int32[](int64)', (6,), [0, -1, -2, -3, -4, -5]),
    ('ne_array', 'bool(int64, int32)', (6, 3), False),
    ('not_in', 'int8(int64, int32)', (6, 3), False),
    ('or_bw', 'int32[](int64)', (6,), [5, 5, 3, 3, 5, 5]),
    ('pos', 'int32[](int64)', (6,), [0, -1, -2, -3, -4, -5]),
    ('pow', 'int32[](int64)', (6,), [1, 32, 81, 64, 25, 6]),
    ('rshift', 'int32[](int64)', (6,), [0, 0, 0, 0, 2, 5]),
    ('sub', 'int32[](int64)', (6,), [-5, -3, -1, 1, 3, 5]),
    ('truediv', 'int32[](int64)', (6,), [3, 2, 2, 2, 2, 1]),
    ('truediv2', 'double[](int64)', (6,), [3.3333333333333335, 2.75, 2.4, 2.1666666666666665, 2.0, 1.875]),  # noqa: E501
    ('xor', 'int32[](int64)', (6,), [5, 5, 1, 1, 5, 5]),
]


def operator_countOf(size, fill_value, b):
    a = Array(size, 'int64')
    for i in range(size):
        a[i] = fill_value
    return operator.countOf(a, b)


def operator_neg(size):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return operator.neg(a)


def operator_abs(size):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(-i)
    return abs(a)


def operator_pos(size):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(-i)
    return operator.pos(a)


def operator_rshift(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    return operator.rshift(a, b)


def operator_irshift(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    operator.irshift(a, b)
    return a


def operator_lshift(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    return operator.lshift(a, b)


def operator_ilshift(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    operator.ilshift(a, b)
    return a


def operator_floordiv(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i+10)
        b[i] = nb_types.int32(i+3)
    return operator.floordiv(a, b)


def operator_ifloordiv(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i+10)
        b[i] = nb_types.int32(i+3)
    operator.ifloordiv(a, b)
    return a


def operator_floordiv2(size):
    a = Array(size, 'double')
    b = Array(size, 'double')
    for i in range(size):
        a[i] = nb_types.double(i+10)
        b[i] = nb_types.double(i+3)
    return operator.floordiv(a, b)


def operator_ifloordiv2(size):
    a = Array(size, 'double')
    b = Array(size, 'double')
    for i in range(size):
        a[i] = nb_types.double(i+10)
        b[i] = nb_types.double(i+3)
    operator.ifloordiv(a, b)
    return a


def operator_truediv(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i+10)
        b[i] = nb_types.int32(i+3)
    return operator.truediv(a, b)


def operator_itruediv(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i+10)
        b[i] = nb_types.int32(i+3)
    operator.itruediv(a, b)
    return a


def operator_truediv2(size):
    a = Array(size, 'double')
    b = Array(size, 'double')
    for i in range(size):
        a[i] = nb_types.double(i+10)
        b[i] = nb_types.double(i+3)
    return operator.truediv(a, b)


def operator_itruediv2(size):
    a = Array(size, 'double')
    b = Array(size, 'double')
    for i in range(size):
        a[i] = nb_types.double(i+10)
        b[i] = nb_types.double(i+3)
    operator.itruediv(a, b)
    return a


def operator_pow(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i+1)
        b[i] = nb_types.int32(size-i)
    return operator.pow(a, b)


def operator_ipow(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i+1)
        b[i] = nb_types.int32(size-i)
    operator.ipow(a, b)
    return a


def operator_mul(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    return operator.mul(a, b)


def operator_imul(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    operator.imul(a, b)
    return a


def operator_mod(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i * 123)
        b[i] = nb_types.int32(7)
    return operator.mod(a, b)


def operator_imod(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i * 123)
        b[i] = nb_types.int32(7)
    operator.imod(a, b)
    return a


def operator_sub(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    return operator.sub(a, b)


def operator_isub(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    operator.isub(a, b)
    return a


def operator_add(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    return operator.add(a, b)


def operator_iadd(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(1)
    operator.iadd(a, b)
    return a


def operator_xor(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    return operator.xor(a, b)


def operator_ixor(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    operator.ixor(a, b)
    return a


def operator_or_bw(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    return operator.or_(a, b)


def operator_ior(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    operator.ior(a, b)
    return a


def operator_and_bw(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    return operator.and_(a, b)


def operator_iand(size):
    a = Array(size, 'int32')
    b = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
        b[i] = nb_types.int32(size-i-1)
    operator.iand(a, b)
    return a


def operator_eq(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a == v


def operator_eq_array(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a == a


def operator_ne(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a != v


def operator_ne_array(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a != a


def operator_lt(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a < v


def operator_lt_array(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a < a


def operator_le(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a <= v


def operator_le_array(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a <= a


def operator_gt(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a > v


def operator_gt_array(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a > a


def operator_ge(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a >= v


def operator_ge_array(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return a >= a


def operator_in(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return v in a


def operator_not_in(size, v):
    a = Array(size, 'int32')
    for i in range(size):
        a[i] = nb_types.int32(i)
    return v not in a


def operator_is(size, v):
    a = Array(size, 'int32')
    a.fill(v)
    return a is a


def operator_is_not(size, v):
    a = Array(size, 'int32')
    a.fill(v)
    return a is not a


def operator_is_not2(size, v):
    a = Array(size, 'int32')
    a.fill(v)
    b = Array(size, 'int32')
    b.fill(v)
    return a is not b


@pytest.mark.parametrize("suffix, signature, args, expected", operator_methods,
                         ids=[item[0] for item in operator_methods])
def test_array_operators(omnisci, suffix, signature, args, expected):

    if omnisci.has_cuda and suffix in ['countOf', 'in', 'not_in'] and omnisci.version < (5, 5):
        # https://github.com/xnd-project/rbc/issues/107
        pytest.skip(f'operator_{suffix}: crashes CUDA enabled omniscidb server'
                    ' [rbc issue 107]')

    if (available_version[:3] == (5, 3, 1)
        and suffix in ['abs', 'add', 'and_bw', 'eq', 'floordiv', 'floordiv2',
                       'ge', 'gt', 'iadd', 'iand', 'ifloordiv', 'ifloordiv2',
                       'ilshift', 'imul', 'ior', 'isub', 'ipow', 'irshift',
                       'itruediv', 'itruediv2', 'imod', 'ixor', 'le', 'lshift',
                       'lt', 'mul', 'mod', 'ne', 'neg', 'or_bw', 'pos', 'pow',
                       'rshift', 'sub', 'truediv', 'truediv2', 'xor']):
        pytest.skip(
            f'operator_{suffix}: crashes CPU-only omniscidb server v 5.3.1'
            ' [issue 115]')

    omnisci.reset()

    omnisci(signature)(eval('operator_{}'.format(suffix)))

    query = 'select operator_{suffix}'.format(**locals()) + \
            '(' + ', '.join(map(str, args)) + ')'
    _, result = omnisci.sql_execute(query)
    out = list(result)[0]

    if suffix in ['in', 'not_in']:
        assert (expected == out[0]), 'operator_' + suffix
    elif '_array' in suffix:
        assert (expected == out[0]), 'operator_' + suffix
    else:
        assert np.array_equal(expected, out[0]), 'operator_' + suffix



"""Collections of library function names.
"""


class Library:
    """Base class for a collection of library function names.
    """

    @staticmethod
    def get(libname, _cache={}):
        if libname in _cache:
            return _cache[libname]
        if libname == 'stdlib':
            r = Stdlib()
        elif libname == 'stdio':
            r = Stdio()
        elif libname == 'm':
            r = Mlib()
        elif libname == 'libdevice':
            r = Libdevice()
        elif libname == 'nvvm':
            r = NVVMIntrinsics()
        elif libname == 'llvm':
            r = LLVMIntrinsics()
        elif libname == 'omniscidb':
            r = OmnisciDB()
        else:
            raise ValueError(f'Unknown library {libname}')
        _cache[libname] = r
        return r

    def __contains__(self, fname):
        return self.check(fname)

    def check(self, fname):
        """
        Return True if library contains a function with given name.
        """
        return fname in self._function_names


class OmnisciDB(Library):

    name = 'omniscidb'

    _function_names = list('''
    allocate_varlen_buffer
    '''.strip().split())


class Stdlib(Library):
    """
    Reference: http://www.cplusplus.com/reference/cstdlib/
    """
    name = 'stdlib'
    _function_names = list(''' atof atoi atol atoll strtod strtof strtol strtold strtoll strtoul
    strtoull rand srand calloc free malloc realloc abort atexit
    at_quick_exit exit getenv quick_exit system bsearch qsort abs div
    labs ldiv llabs lldiv mblen mbtowc wctomb mbstowcs wcstombs '''.strip().split())


class Stdio(Library):
    """
    Reference: http://www.cplusplus.com/reference/cstdio/
    """
    name = 'stdio'

    _function_names = list(''' remove rename tmpfile tmpnam fclose fflush fopen freopen setbuf
    setvbuf fprintf fscanf printf scanf snprintf sprintf sscanf
    vfprintf vfscanf vprintf vscanf vsnprintf vsprintf vsscanf fgetc
    fgets fputc fputs getc getchar gets putc putchar puts ungetc fread
    fwrite fgetpos fseek fsetpos ftell rewind clearerr feof ferror
    perror '''.strip().split())


class Mlib(Library):
    """
    References:
      https://www.gnu.org/software/libc/manual/html_node/Mathematics.html
      https://en.cppreference.com/w/cpp/header/cmath
    """
    name = 'm'

    _function_names = list('''sin sinf sinl cos cosf cosl tan tanf tanl sincos sincosf sincosl
    csin csinf csinl ccos ccosf ccosl ctan ctanf ctanl asin asinf
    asinl acos acosf acosl atan atanf atanl atan2 atan2f atan2l casin
    casinf casinl cacos cacosf cacosl catan catanf catanl exp expf
    expl exp2 exp2f exp2l exp10 exp10f exp10l log logf logl log2 log2f
    log2l log10 log10f log10l logb logbf logbl ilogb ilogbf ilogbl pow
    powf powl sqrt sqrtf sqrtl cbrt cbrtf cbrtl hypot hypotf hypotl
    expm1 expm1f expm1l log1p log1pf log1pl clog clogf clogl clog10
    clog10f clog10l csqrt csqrtf csqrtl cpow cpowf cpowl sinh sinhf
    sinhl cosh coshf coshl tanh tanhf tanhl csinh csinhf csinhl ccosh
    ccoshf ccoshl ctanh ctanhf ctanhl asinh asinhf asinhl acosh acoshf
    acoshl atanh atanhf atanhl casinh casinhf casinhl cacosh cacoshf
    cacoshl catanh catanhf catanhl erf erff erfl erfc erfcf erfcl
    lgamma lgammaf lgammal tgamma tgammaf tgammal lgamma_r lgammaf_r
    lgammal_r gamma gammaf gammal j0 j0f j0l j1 j1f j1l jn jnf jnl y0
    y0f y0l y1 y1f y1l yn ynf ynl rand srand rand_r random srandom
    initstate setstate random_r srandom_r initstate_r setstate_r
    drand48 erand48 lrand48 nrand48 mrand48 jrand48 srand48 seed48
    lcong48 drand48_r erand48_r lrand48_r nrand48_r mrand48_r
    jrand48_r srand48_r seed48_r lcong48_r abs labs llabs fabs fabsf
    fabsl cabs cabsf cabsl frexp frexpf frexpl ldexp ldexpf ldexpl
    scalb scalbf scalbl scalbn scalbnf scalbnl significand
    significandf significandl ceil ceilf ceill floor floorf floorl
    trunc truncf truncl rint rintf rintl nearbyint nearbyintf
    nearbyintl round roundf roundl roundeven roundevenf roundevenl
    lrint lrintf lrintl lround lroundf lroundl llround llroundf
    llroundl fromfp fromfpf fromfpl ufromfp ufromfpf ufromfpl fromfpx
    fromfpxf fromfpxl ufromfpx ufromfpxf ufromfpxl modf modff modfl
    fmod fmodf fmodl remainder remainderf remainderl drem dremf dreml
    copysign copysignf copysignl signbit signbitf signbitl nextafter
    nextafterf nextafterl nexttoward nexttowardf nexttowardl nextup
    nextupf nextupl nextdown nextdownf nextdownl nan nanf nanl
    canonicalize canonicalizef canonicalizel getpayload getpayloadf
    getpayloadl setpayload setpayloadf setpayloadl setpayloadsig
    setpayloadsigf setpayloadsigl isgreater isgreaterequal isless
    islessequal islessgreater isunordered iseqsig totalorder
    totalorderf totalorderl totalordermag totalorderf totalorderl fmin
    fminf fminl fmax fmaxf fmaxl fminmag fminmagf fminmagl fmaxmag
    fmaxmagf fmaxmagl fdim fdimf fdiml fma fmaf fmal fadd faddf faddl
    fsub fsubf fsubl fmul fmulf fmull fdiv fdivf fdivl '''.strip().split())


def drop_suffix(f):
    s = f.rsplit('.', 1)[-1]
    if s in ['p0i8', 'f64', 'f32', 'i1', 'i8', 'i16', 'i32', 'i64', 'i128']:
        f = f[:-len(s)-1]
        return drop_suffix(f)
    return f


def get_llvm_name(f, prefix='llvm.'):
    """Return normalized name of a llvm intrinsic name.
    """
    if f.startswith(prefix):
        return drop_suffix(f[len(prefix):])
    return f


class LLVMIntrinsics(Library):
    """LLVM intrinsic function names with prefix `llvm.` removed.

    Reference: https://llvm.org/docs/LangRef.html#intrinsic-functions
    """
    name = 'llvm'

    def check(self, fname):
        if fname.startswith('llvm.'):
            return Library.check(self, get_llvm_name(fname))
        return False

    _function_names = list(''' va_start va_end va_copy gcroot gcread gcwrite returnaddress
    addressofreturnaddress sponentry frameaddress stacksave
    stackrestore get.dynamic.area.offset prefetch pcmarker
    readcyclecounter clear_cache instrprof.increment
    instrprof.increment.step instrprof.value.profile thread.pointer
    call.preallocated.setup call.preallocated.arg
    call.preallocated.teardown abs smax smin umax umin memcpy
    memcpy.inline memmove sqrt powi sin cos pow exp exp2 log log10
    log2 fma fabs minnum maxnum minimum maximum copysign floor ceil
    trunc rint nearbyint round roundeven lround llround lrint llrint
    ctpop ctlz cttz fshl fshr sadd.with.overflow uadd.with.overflow
    ssub.with.overflow usub.with.overflow smul.with.overflow
    umul.with.overflow sadd.sat uadd.sat ssub.sat usub.sat sshl.sat
    ushl.sat smul.fix umul.fix smul.fix.sat umul.fix.sat sdiv.fix
    udiv.fix sdiv.fix.sat udiv.fix.sat canonicalize fmuladd
    set.loop.iterations test.set.loop.iterations loop.decrement.reg
    loop.decrement vector.reduce.add vector.reduce.fadd
    vector.reduce.mul vector.reduce.fmul vector.reduce.and
    vector.reduce.or vector.reduce.xor vector.reduce.smax
    vector.reduce.smin vector.reduce.umax vector.reduce.umin
    vector.reduce.fmax vector.reduce.fmin matrix.transpose
    matrix.multiply matrix.column.major.load matrix.column.major.store
    convert.to.fp16 convert.from.fp16 init.trampoline
    adjust.trampoline lifetime.start lifetime.end invariant.start
    invariant.end launder.invariant.group strip.invariant.group
    experimental.constrained.fadd experimental.constrained.fsub
    experimental.constrained.fmul experimental.constrained.fdiv
    experimental.constrained.frem experimental.constrained.fma
    experimental.constrained.fptoui experimental.constrained.fptosi
    experimental.constrained.uitofp experimental.constrained.sitofp
    experimental.constrained.fptrunc experimental.constrained.fpext
    experimental.constrained.fmuladd experimental.constrained.sqrt
    experimental.constrained.pow experimental.constrained.powi
    experimental.constrained.sin experimental.constrained.cos
    experimental.constrained.exp experimental.constrained.exp2
    experimental.constrained.log experimental.constrained.log10
    experimental.constrained.log2 experimental.constrained.rint
    experimental.constrained.lrint experimental.constrained.llrint
    experimental.constrained.nearbyint experimental.constrained.maxnum
    experimental.constrained.minnum experimental.constrained.maximum
    experimental.constrained.minimum experimental.constrained.ceil
    experimental.constrained.floor experimental.constrained.round
    experimental.constrained.roundeven experimental.constrained.lround
    experimental.constrained.llround experimental.constrained.trunc
    flt.rounds var.annotation ptr.annotation annotation
    codeview.annotation trap debugtrap stackprotector stackguard
    objectsize expect expect.with.probability assume ssa_copy
    type.test type.checked.load donothing experimental.deoptimize
    experimental.guard experimental.widenable.condition load.relative
    sideeffect is.constant ptrmask vscale
    memcpy.element.unordered.atomic memmove.element.unordered.atomic
    memset.element.unordered.atomic objc.autorelease
    objc.autoreleasePoolPop objc.autoreleasePoolPush
    objc.autoreleaseReturnValue objc.copyWeak objc.destroyWeak
    objc.initWeak objc.loadWeak objc.loadWeakRetained objc.moveWeak
    objc.release objc.retain objc.retainAutorelease
    objc.retainAutoreleaseReturnValue
    objc.retainAutoreleasedReturnValue objc.retainBlock
    objc.storeStrong objc.storeWeak preserve.array.access.index
    preserve.union.access.index preserve.struct.access.index '''.strip().split())


class NVVMIntrinsics(Library):
    """NVVM intrinsic function names with prefix `llvm.` removed.

    Reference: https://docs.nvidia.com/cuda/nvvm-ir-spec/index.html#intrinsic-functions
    """
    name = 'nvvm'

    def check(self, fname):
        if fname.startswith('llvm.'):
            return Library.check(self, get_llvm_name(fname))
        return False

    _function_names = list(''' memcpy memmove memset sqrt fma bswap ctpop ctlz cttz fmuladd
    convert.to.fp16.f32 convert.from.fp16.f32 convert.to.fp16
    convert.from.fp16 lifetime.start lifetime.end invariant.start
    invariant.end var.annotation ptr.annotation annotation expect
    donothing '''.strip().split())


class Libdevice(Library):
    """NVIDIA libdevice function names with prefix `__nv_` removed.

    Reference: https://docs.nvidia.com/cuda/libdevice-users-guide/function-desc.html#function-desc
    """
    name = 'libdevice'

    def check(self, fname):
        if fname.startswith('__nv_'):
            return Library.check(self, get_llvm_name(fname, prefix='__nv_'))
        return False

    _function_names = list(''' abs acos acosf acosh acoshf asin asinf asinh asinhf atan atan2
    atan2f atanf atanh atanhf brev brevll byte_perm cbrt cbrtf ceil
    ceilf clz clzll copysign copysignf cos cosf cosh coshf cospi
    cospif dadd_rd dadd_rn dadd_ru dadd_rz ddiv_rd ddiv_rn ddiv_ru
    ddiv_rz dmul_rd dmul_rn dmul_ru dmul_rz double2float_rd
    double2float_rn double2float_ru double2float_rz double2hiint
    double2int_rd double2int_rn double2int_ru double2int_rz
    double2ll_rd double2ll_rn double2ll_ru double2ll_rz double2loint
    double2uint_rd double2uint_rn double2uint_ru double2uint_rz
    double2ull_rd double2ull_rn double2ull_ru double2ull_rz
    double_as_longlong drcp_rd drcp_rn drcp_ru drcp_rz dsqrt_rd
    dsqrt_rn dsqrt_ru dsqrt_rz erf erfc erfcf erfcinv erfcinvf erfcx
    erfcxf erff erfinv erfinvf exp exp10 exp10f exp2 exp2f expf expm1
    expm1f fabs fabsf fadd_rd fadd_rn fadd_ru fadd_rz fast_cosf
    fast_exp10f fast_expf fast_fdividef fast_log10f fast_log2f
    fast_logf fast_powf fast_sincosf fast_sinf fast_tanf fdim fdimf
    fdiv_rd fdiv_rn fdiv_ru fdiv_rz ffs ffsll finitef float2half_rn
    float2int_rd float2int_rn float2int_ru float2int_rz float2ll_rd
    float2ll_rn float2ll_ru float2ll_rz float2uint_rd float2uint_rn
    float2uint_ru float2uint_rz float2ull_rd float2ull_rn float2ull_ru
    float2ull_rz float_as_int floor floorf fma fma_rd fma_rn fma_ru
    fma_rz fmaf fmaf_rd fmaf_rn fmaf_ru fmaf_rz fmax fmaxf fmin fminf
    fmod fmodf fmul_rd fmul_rn fmul_ru fmul_rz frcp_rd frcp_rn frcp_ru
    frcp_rz frexp frexpf frsqrt_rn fsqrt_rd fsqrt_rn fsqrt_ru fsqrt_rz
    fsub_rd fsub_rn fsub_ru fsub_rz hadd half2float hiloint2double
    hypot hypotf ilogb ilogbf int2double_rn int2float_rd int2float_rn
    int2float_ru int2float_rz int_as_float isfinited isinfd isinff
    isnand isnanf j0 j0f j1 j1f jn jnf ldexp ldexpf lgamma lgammaf
    ll2double_rd ll2double_rn ll2double_ru ll2double_rz ll2float_rd
    ll2float_rn ll2float_ru ll2float_rz llabs llmax llmin llrint
    llrintf llround llroundf log log10 log10f log1p log1pf log2 log2f
    logb logbf logf longlong_as_double max min modf modff mul24
    mul64hi mulhi nan nanf nearbyint nearbyintf nextafter nextafterf
    normcdf normcdff normcdfinv normcdfinvf popc popcll pow powf powi
    powif rcbrt rcbrtf remainder remainderf remquo remquof rhadd rint
    rintf round roundf rsqrt rsqrtf sad saturatef scalbn scalbnf
    signbitd signbitf sin sincos sincosf sincospi sincospif sinf sinh
    sinhf sinpi sinpif sqrt sqrtf tan tanf tanh tanhf tgamma tgammaf
    trunc truncf uhadd uint2double_rn uint2float_rd uint2float_rn
    uint2float_ru uint2float_rz ull2double_rd ull2double_rn
    ull2double_ru ull2double_rz ull2float_rd ull2float_rn ull2float_ru
    ull2float_rz ullmax ullmin umax umin umul24 umul64hi umulhi urhadd
    usad y0 y0f y1 y1f yn ynf '''.strip().split())



"""
rbc-specific errors and warnings.
"""

import numba
from .utils import get_version

if get_version('numba') >= (0, 49):
    numba_errors = numba.core.errors
else:
    numba_errors = numba.errors


class OmnisciServerError(Exception):
    """
    Launch when OmnisciDB server raises a runtime error that RBC knows
    how to interpret.
    """


class UnsupportedError(numba_errors.UnsupportedError):
    """
    Launch when an attempt is to use a feature that is not supported
    for a given target.
    """


class ForbiddenNameError(Exception):
    """
    Launch when the user defines a function with name
    in a blacklist. For more info, see:
    https://github.com/xnd-project/rbc/issues/32
    """
    pass


class ForbiddenIntrinsicError(Exception):
    """
    Raised when the user uses an intrinsic with name
    in a blacklist. For more info, see:
    https://github.com/xnd-project/rbc/issues/207
    """
    pass
